from typing import Iterable, Union

from packaging.version import Version

import redgrease
import redgrease.runtime


class GearReader(redgrease.runtime.GearsBuilder):
    def __init__(
        self,
        reader: str,
        defaultArg: str = "*",
        desc: str = None,
        requirements: Iterable[str] = None,
        require_runtime: Union[Version, str, bool] = True,
    ):
        """Gear function / process factory
        Args:
            reader (str, optional): Input records reader
                Defining where the input to the gear will come from.
                One of:
                - 'KeysReader':
                - 'KeysOnlyReader':
                - 'StreamReader':
                - 'PythonReader':
                - 'ShardsReader':
                - 'CommandReader':
                Defaults to 'KeysReader'.

            defaultArg (str, optional):
                Additional arguments to the reader.
                These are usually a key's name, prefix, glob-like
                or a regular expression.
                Its use depends on the function's reader type and action.
                Defaults to '*'.

            desc (str, optional): An optional description.
                Defaults to None.

            requirements (Iterable[str], optional):
                Package lependencies for the gear train.
                Defaults to None.

            require_runtime (Union[Version, str, bool], optional):
                Auto-require the redgrease server runtime.
                Defaults to False.
        """
        super().__init__(reader=reader, defaultArg=defaultArg, desc=desc)

        if require_runtime:
            if isinstance(require_runtime, Version):
                runtime_package = f"redgrease[runtime]=={require_runtime}"
            elif isinstance(require_runtime, str):
                if require_runtime[0] in ["=", ">", "<", "!"]:
                    runtime_package = f"redgrease[runtime]{require_runtime}"
                elif require_runtime[0].isnumeric():
                    runtime_package = f"redgrease[runtime]=={require_runtime}"
                else:
                    runtime_package = require_runtime
            else:
                runtime_package = "redgrease[runtime]"
            self.requirements = [runtime_package]
        else:
            self.requirements = []

        if requirements:
            self.requirements += list(requirements)


class KeysReader(GearReader):
    def __init__(
        self,
        default_key_pattern: str = "*",
    ):
        super().__init__(
            reader=redgrease.Reader.KeysReader, defaultArg=default_key_pattern
        )
        self.default_key_pattern = default_key_pattern


class KeysOnlyReader(GearReader):
    def __init__(
        self,
        default_key_pattern: str = "*",
    ):
        super().__init__(
            reader=redgrease.Reader.KeysOnlyReader, defaultArg=default_key_pattern
        )
        self.default_key_pattern = default_key_pattern


class StreamReader(GearReader):
    def __init__(
        self,
        default_key_pattern: str = "*",
    ):
        super().__init__(
            reader=redgrease.Reader.StreamReader, defaultArg=default_key_pattern
        )
        self.default_key_pattern = default_key_pattern


class PythonReader(GearReader):
    def __init__(self):
        super().__init__(reader=redgrease.Reader.PythonReader)


class ShardsIDReader(GearReader):
    def __init__(self):
        super().__init__(reader=redgrease.Reader.ShardsIDReader)


class CommandReader(GearReader):
    def __init__(self):
        super().__init__(reader=redgrease.Reader.CommandReaderReader)



import logging
from typing import Any, Dict, List

import redis

from redgrease.typing import Constructor
from redgrease.utils import safe_str, to_list

log = logging.getLogger(__name__)


class Config:
    ValueTypes: Dict[str, Constructor[Any]] = {
        "MaxExecutions": int,
        "MaxExecutionsPerRegistration": int,
        "ProfileExecutions": bool,
        "PythonAttemptTraceback": bool,
        "DownloadDeps": bool,
        "DependenciesUrl": safe_str,
        "DependenciesSha256": safe_str,
        "PythonInstallationDir": safe_str,
        "CreateVenv": bool,
        "ExecutionThreads": int,
        "ExecutionMaxIdleTime": int,
        "PythonInstallReqMaxIdleTime": int,
        "SendMsgRetries": int,
    }

    __slots__ = "redis"

    def __init__(self, redis: redis.Redis):
        self.redis = redis

    def get(self, *config_option) -> List[str]:
        """Get the value of one or more built-in configuration or
        a user-defined options.

        Args:
            config_name (Union[ID, bytes, str]): One or more names/key
            of configurations to get

        Returns:
            List of config values
        """
        if config_option:
            conf = [safe_str(c) for c in config_option]
        else:
            conf = list(Config.ValueTypes.keys())

        return self.redis.execute_command("RG.CONFIGGET", *conf, keys=conf)

    def set(self, config_dict=None, **config_setting) -> bool:
        """Set a value of one ore more built-in configuration or
        a user-defined options.

        Args:
            config_setting: Key-value-pairs of config settings

        Returns:
            bool: True if all was successful, false oterwise
        """
        settings = {}
        if config_dict:
            settings.update(config_dict)

        if config_setting:
            settings.update(config_setting)

        return self.redis.execute_command("RG.CONFIGSET", *to_list(settings))

    def get_single(self, key):
        return self.get(key)[key]

    @property
    def MaxExecutions(self):
        return self.get_single("MaxExecutions")

    @MaxExecutions.setter
    def MaxExecutions(self, value):
        return self.set(MaxExecutions=value)

    @property
    def MaxExecutionsPerRegistration(self):
        return self.get_single("MaxExecutionsPerRegistration")

    @MaxExecutionsPerRegistration.setter
    def MaxExecutionsPerRegistration(self, value):
        self.set(MaxExecutionsPerRegistration=value)

    @property
    def ProfileExecutions(self):
        return self.get_single("ProfileExecutions")

    @ProfileExecutions.setter
    def ProfileExecutions(self, value):
        self.set(ProfileExecutions=value)

    @property
    def PythonAttemptTraceback(self):
        return self.get_single("PythonAttemptTraceback")

    @PythonAttemptTraceback.setter
    def PythonAttemptTraceback(self, value):
        self.set(PythonAttemptTraceback=value)

    @property
    def DownloadDeps(self):
        return self.get_single("DownloadDeps")

    # @DownloadDeps.setter
    # def DownloadDeps(self, value):
    #     self.set(DownloadDeps=value)

    @property
    def DependenciesUrl(self):
        return self.get_single("DependenciesUrl")

    # @DependenciesUrl.setter
    # def DependenciesUrl(self, value):
    #     self.set(DependenciesUrl=value)

    @property
    def DependenciesSha256(self):
        return self.get_single("DependenciesSha256")

    # @DependenciesSha256.setter
    # def DependenciesSha256(self, value):
    #     self.set(DependenciesSha256=value)

    @property
    def PythonInstallationDir(self):
        return self.get_single("PythonInstallationDir")

    # @PythonInstallationDir.setter
    # def PythonInstallationDir(self, value):
    #     self.set(PythonInstallationDir=value)

    @property
    def CreateVenv(self):
        return self.get_single("CreateVenv")

    # @CreateVenv.setter
    # def CreateVenv(self, value):
    #     self.set(CreateVenv=value)

    @property
    def ExecutionThreads(self):
        return self.get_single("ExecutionThreads")

    # @ExecutionThreads.setter
    # def ExecutionThreads(self, value):
    #     self.set(ExecutionThreads=value)

    @property
    def ExecutionMaxIdleTime(self):
        return self.get_single("ExecutionMaxIdleTime")

    @ExecutionMaxIdleTime.setter
    def ExecutionMaxIdleTime(self, value):
        self.set(ExecutionMaxIdleTime=value)

    @property
    def PythonInstallReqMaxIdleTime(self):
        return self.get_single("PythonInstallReqMaxIdleTime")

    @PythonInstallReqMaxIdleTime.setter
    def PythonInstallReqMaxIdleTime(self, value):
        self.set(PythonInstallReqMaxIdleTime=value)

    @property
    def SendMsgRetries(self):
        return self.get_single("SendMsgRetries")

    @SendMsgRetries.setter
    def SendMsgRetries(self, value):
        self.set(SendMsgRetries=value)



import types
from functools import lru_cache

import redgrease
from redgrease.client import RedisGears


def runtime_execute_command(self, *args, **options):
    "Execute a command in local Gear runtime and return a parsed response"
    command_name = args[0]
    response = redgrease.execute(*args)
    if command_name in self.response_callbacks:
        return self.response_callbacks[command_name](response, **options)
    return response


@lru_cache()
def get_runtime_client():
    runtime_client = RedisGears(connection_pool=...)
    runtime_client.execute_command = types.MethodType(
        runtime_execute_command, runtime_client
    )
    return runtime_client


redis = get_runtime_client()  # noqa: F811 - Used as exported var



from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Dict, Optional, Type

if TYPE_CHECKING:
    from redgrease.gears import PartialGearFunction
    import redgrease.typing as optype

from redgrease.sugar import TriggerMode


@dataclass
class Operation:
    pass

    def add_to(self, function):
        raise NotImplementedError("Cannot add genric operation to Gear function")


@dataclass
class Reader(Operation):
    reader: str
    defaultArg: str
    desc: Optional[str]

    def add_to(self, builder: Type):
        return builder(self.reader, self.defaultArg, self.desc)


@dataclass
class Run(Operation):
    arg: Optional[str] = None
    convertToStr: bool = True
    collect: bool = True
    kwargs: Dict[str, Any] = field(default_factory=dict)

    def add_to(self, function: "PartialGearFunction"):
        import cloudpickle

        return function.map(lambda x: cloudpickle.dumps(x, protocol=4)).run(
            self.arg, False, self.collect, **self.kwargs
        )


@dataclass
class Register(Operation):
    prefix: str = "*"
    convertToStr: bool = True
    collect: bool = True
    mode: str = TriggerMode.Async
    onRegistered: Optional["optype.Callback"] = None
    trigger: Optional[str] = None
    kwargs: Dict[str, Any] = field(default_factory=dict)

    def add_to(self, function: "PartialGearFunction"):
        return function.register(
            self.prefix,
            self.convertToStr,
            self.collect,
            self.mode,
            self.onRegistered,
            self.trigger,
            **self.kwargs
        )


@dataclass
class Map(Operation):
    op: "optype.Mapper"

    def add_to(self, function: "PartialGearFunction"):
        return function.map(self.op)


@dataclass
class FlatMap(Operation):
    op: "optype.Expander"

    def add_to(self, function: "PartialGearFunction"):
        return function.flatmap(self.op)


@dataclass
class ForEach(Operation):
    op: "optype.Processor"

    def add_to(self, function: "PartialGearFunction"):
        return function.foreach(self.op)


@dataclass
class Filter(Operation):
    op: "optype.Filterer"

    def add_to(self, function: "PartialGearFunction"):
        return function.filter(self.op)


@dataclass
class Accumulate(Operation):
    op: "optype.Accumulator"

    def add_to(self, function: "PartialGearFunction"):
        return function.accumulate(self.op)


@dataclass
class LocalGroupBy(Operation):
    extractor: "optype.Extractor"
    reducer: "optype.Reducer"

    def add_to(self, function: "PartialGearFunction"):
        return function.localgroupby(self.extractor, self.reducer)


@dataclass
class Limit(Operation):
    length: int
    start: int

    def add_to(self, function: "PartialGearFunction"):
        return function.limit(self.length, self.start)


@dataclass
class Collect(Operation):
    pass

    def add_to(self, function: "PartialGearFunction"):
        return function.collect()


@dataclass
class Repartition(Operation):
    extractor: "optype.Extractor"

    def add_to(self, function: "PartialGearFunction"):
        return function.repartition(self.extractor)


@dataclass
class Aggregate(Operation):
    zero: Any
    seqOp: "optype.Accumulator"
    combOp: "optype.Accumulator"

    def add_to(self, function: "PartialGearFunction"):
        return function.aggregate(self.zero, self.seqOp, self.combOp)


@dataclass
class AggregateBy(Operation):
    extractor: "optype.Extractor"
    zero: Any
    seqOp: "optype.Reducer"
    combOp: "optype.Reducer"

    def add_to(self, function: "PartialGearFunction"):
        return function.aggregateby(self.extractor, self.zero, self.seqOp, self.combOp)


@dataclass
class GroupBy(Operation):
    extractor: "optype.Extractor"
    reducer: "optype.Reducer"

    def add_to(self, function: "PartialGearFunction"):
        return function.groupby(self.extractor, self.reducer)


@dataclass
class BatchGroupBy(Operation):
    extractor: "optype.Extractor"
    reducer: "optype.BatchReducer"

    def add_to(self, function: "PartialGearFunction"):
        return function.batchgroupby(self.extractor, self.reducer)


@dataclass
class Sort(Operation):
    reverse: bool = True

    def add_to(self, function: "PartialGearFunction"):
        return function.sort(self.reverse)


@dataclass
class Distinct(Operation):
    pass

    def add_to(self, function: "PartialGearFunction"):
        return function.distinct()


@dataclass
class Count(Operation):
    pass

    def add_to(self, function: "PartialGearFunction"):
        return function.count()


@dataclass
class CountBy(Operation):
    extractor: "optype.Extractor"

    def add_to(self, function: "PartialGearFunction"):
        return function.countby(self.extractor)


@dataclass
class Avg(Operation):
    extractor: "optype.Extractor"

    def add_to(self, function: "PartialGearFunction"):
        return function.avg(self.extractor)



from typing import Any, Callable, Dict, Iterable, Type, TypeVar, Union

try:
    from typing import Protocol
except ImportError:
    from typing_extensions import Protocol  # type: ignore

# Type aliases for type hints

T = TypeVar("T")

InputRecord = TypeVar("InputRecord", contravariant=True)
OutputRecord = TypeVar("OutputRecord", covariant=True)

Key = TypeVar("Key", contravariant=True)
Val = TypeVar("Val")

Constructor = Union[Type[T], Callable[[Any], T]]

RedisType = Union[bytes, int, float]
SafeType = Union[bytes, int, float, str]
SupportedType = Union[bool, str, bytes, int, float]

RedisKey = Union[str, bytes]

Record = Dict


class Callback(Protocol[InputRecord]):
    def __call__(self, record: InputRecord) -> Any:
        ...


class Extractor(Protocol[InputRecord]):
    def __call__(self, record: InputRecord) -> str:
        ...


class Mapper(Protocol[InputRecord, OutputRecord]):
    def __call__(self, record: InputRecord) -> OutputRecord:
        ...


class Expander(Protocol[InputRecord, OutputRecord]):
    def __call__(self, record: InputRecord) -> Iterable[OutputRecord]:
        ...


class Processor(Protocol[InputRecord]):
    def __call__(self, record: InputRecord) -> None:
        ...


class Filterer(Protocol[InputRecord]):
    def __call__(self, record: InputRecord) -> bool:
        ...


class Accumulator(Protocol[T, InputRecord]):
    def __call__(self, accumulator: T, record: InputRecord) -> T:
        ...


class Reducer(Protocol[Key, T, InputRecord]):
    def __call__(self, key: Key, accumulator: T, record: InputRecord) -> T:
        ...


class BatchReducer(Protocol[Key, OutputRecord, InputRecord]):
    def __call__(self, key: Key, records: Iterable[InputRecord]) -> OutputRecord:
        ...
